/*
	basicConsumer.java - tutorial6 - Request & Decode Machine Readable News

	This tutorial is an independent part in a series of tutorials with the goal of creating an advanced consumer application.

	The goal of this tutorial is to extend Tutorial 3 to request, retrieve and decode News Text Analytics (NTA)
    domain from a Machine Readable News (MRN) service.

	This tutorial is the sixth part in a series of tutorials with the goal of creating a basic consumer application.

	The series includes the following:

		o Tutorial 1 - Creating a barebones EMA consumer application
		o Tutorial 2 - Requesting and displaying MarketPrice data
		o Tutorial 3 - Parsing and Decoding MarketPrice data
		o Tutorial 4 - Requesting, parsing and decoding Level 2 data
		o Tutorial 5 - Dispatching messages within the user thread
		o Tutorial 6 - Request & Decode Machine Readable News
		
	For an example code described in the Developer Portal site, please refer to a Consumer.java in the same tutorial6 package.
	
 */

package com.thomsonreuters.ema.tutorials.consumer.tutorial6;

import java.io.*;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.zip.GZIPInputStream;

import com.thomsonreuters.ema.access.AckMsg;
import com.thomsonreuters.ema.access.Data;
import com.thomsonreuters.ema.access.DataType;
import com.thomsonreuters.ema.access.DataType.DataTypes;
import com.thomsonreuters.ema.access.EmaFactory;
import com.thomsonreuters.ema.access.EmaUtility;
import com.thomsonreuters.ema.access.FieldEntry;
import com.thomsonreuters.ema.access.FieldList;
import com.thomsonreuters.ema.access.GenericMsg;
import com.thomsonreuters.ema.access.Map;
import com.thomsonreuters.ema.access.MapEntry;
import com.thomsonreuters.ema.access.Msg;
import com.thomsonreuters.ema.access.OmmConsumer;
import com.thomsonreuters.ema.access.OmmConsumerClient;
import com.thomsonreuters.ema.access.OmmConsumerConfig;
import com.thomsonreuters.ema.access.OmmConsumerEvent;
import com.thomsonreuters.ema.access.OmmException;
import com.thomsonreuters.ema.access.Payload;
import com.thomsonreuters.ema.access.RefreshMsg;
import com.thomsonreuters.ema.access.StatusMsg;
import com.thomsonreuters.ema.access.UpdateMsg;
import com.thomsonreuters.ema.rdm.EmaRdm;

// basicConsumer
// The basicConsumer interface manages all control for the consumption of real-time data.  Implementing the
// EMA interface: OmmConsumerClient, we now have the opportunity to capture multiple, realtime events to
// process the different types of messages we can receive once we subscribe to a market data item.  The
// interface below implements a number of required callbacks to achieve this goal.
//
public class basicConsumer implements OmmConsumerClient
{
    // Define our consumer communication channel to our Provider managing connection and item registration
    OmmConsumer consumer = null;

    //*************************************************************************************************
    // Callback methods required by the OmmConsumerClient Interface.
    //
    // The following methods allow us to capture realtime events such as status and data events.
    //*************************************************************************************************

    // onRefreshMsg
    // Capture the initial image after subscribing (registering interest) in a market data item.  This
    // callback parses and decodes the message and displays to the terminal.
    public void onRefreshMsg(RefreshMsg refreshMsg, OmmConsumerEvent event)
    {
        System.out.println("RefreshMsg");

        // Display the Service Name and Item Name
        // These two fields are typically sent in a RefreshMsg from the Provider.  However, this information
        // is usually not sent for subsequent updates (UpdateMsg), unless explicitely requested during
        // registration.  As a result, we first check of their existence as a safeguard.

        // Service Name
        String serviceName = "Not Available";
        if (refreshMsg.hasServiceName())
        {
            serviceName = refreshMsg.serviceName();
        }
        System.out.println("\tService Name: " + serviceName);

        // Item Name
        String itemName = "Not Available";
        if (refreshMsg.hasName())
        {
            itemName = refreshMsg.name();
        }
        System.out.println("\tItem: " + itemName);

        // Retrieve and display the state of the item. In this example we directly print the item state.
        // However, finer grained information can be retrieved via the OmmState methods.
        System.out.println("\tState: " + refreshMsg.state());

        if ( DataType.DataTypes.MAP == refreshMsg.payload().dataType() )
            decode( refreshMsg.payload().map() );
        else if ( DataType.DataTypes.FIELD_LIST == refreshMsg.payload().dataType() )
            decodePrint( refreshMsg.payload().fieldList() );
        System.out.println();
    }

    // onUpdateMsg
    // After receiving the initial image, we capture any updates to the item based on market events.
    // This callback parses and decodes the message and displays to the terminal.
    public void onUpdateMsg(UpdateMsg updateMsg, OmmConsumerEvent event)
    {
        System.out.println("UpdateMsg");

        // Decode the Payload
        // The data for an update message is conveyed by the payload. As with the refresh message, payload
        // will contain a complex data structure that must be decoded and parsed to retrieve the individual
        // fields of interest. However, the updatehMsg will contain only those fields that have changed
        // based on the market conditions.
        if ( DataType.DataTypes.MAP == updateMsg.payload().dataType() )
            decode( updateMsg.payload().map() );
        else if ( DataType.DataTypes.FIELD_LIST == updateMsg.payload().dataType() )
            decode( updateMsg.payload().fieldList() );
        System.out.println();
    }

    // onStatusMsg
    // General status messages related to your consumer item registration.  For example, you will receive a
    // status message if you register interest in an invalid item.  This callback utilizes the convenient
    // print method to display output.
    public void onStatusMsg(StatusMsg statusMsg, OmmConsumerEvent event)
    {
        System.out.println(statusMsg);
    }

    // onAllMsg, onAckMsg, onGenericMsg
    // These callbacks are not necessary for our series of tutorials.  You can refer to the documentation
    // to better understand the purpose of these callbacks.
    public void onAllMsg(Msg msg, OmmConsumerEvent consumerEvent){}
    public void onAckMsg(AckMsg ackMsg, OmmConsumerEvent consumerEvent){}
    public void onGenericMsg(GenericMsg genericMsg, OmmConsumerEvent consumerEvent){}


    //*************************************************************************************************
    // Helpers and other private methods.
    //
    //*************************************************************************************************
    void decode( Map map )
    {
        System.out.println( "Decode map:");
        if ( DataTypes.FIELD_LIST == map.summaryData().dataType() )
        {
            System.out.println( "Map Summary data:" );
            decode( map.summaryData().fieldList() );
            System.out.println();
        }

        Iterator<MapEntry> iter = map.iterator();
        MapEntry mapEntry;
        while ( iter.hasNext() )
        {
            mapEntry = iter.next();

            if ( DataTypes.BUFFER == mapEntry.key().dataType() )
                System.out.println( "Action: " + mapEntry.mapActionAsString() + " key value: " +
                        EmaUtility.asHexString(mapEntry.key().buffer().buffer()));

            if ( DataTypes.FIELD_LIST == mapEntry.loadType() )
            {
                System.out.println( "Entry data:" );
                decode( mapEntry.fieldList() );
                System.out.println();
            }
        }
    }

    void decodePrint(FieldList fieldList) {
        boolean retVal;
        System.out.println("Decode fieldList:");
        Iterator<FieldEntry> iter = fieldList.iterator();
        FieldEntry fieldEntry;
        long totalSize = 0;
        String guid = "";
//		String mrn_src = "";
        while (iter.hasNext()) {
            fieldEntry = iter.next();
            System.out.println("FIELD_ENTRY: " + fieldEntry.fieldId() + "/" + fieldEntry.name() + ": " + fieldEntry.load() + " loadType: " + fieldEntry.loadType());

        }
    }


    private ByteBuffer clone(ByteBuffer original) {
        ByteBuffer clone = ByteBuffer.allocate(original.capacity());
        original.rewind();//copy from the beginning
        clone.put(original);
        original.rewind();
        clone.flip();
        return clone;
    }


    void decode(FieldList fieldList) {
        boolean retVal;
        System.out.println("Decode fieldList:");
        Iterator<FieldEntry> iter = fieldList.iterator();
        FieldEntry fieldEntry;
        long totalSize = 0;
        String guid = "";
//		String mrn_src = "";
        while (iter.hasNext()) {
            fieldEntry = iter.next();
            System.out.println("FIELD_ENTRY: " + fieldEntry.fieldId() + "/" + fieldEntry.name() + ": " + fieldEntry.load() + " loadType: " + fieldEntry.loadType());
            if (fieldEntry.loadType() == DataTypes.BUFFER) {

                ByteBuffer bt = clone(fieldEntry.buffer().buffer());

                if (fieldEntry.fieldId() == FID_FRAGMENT) {
                    // System.out.println("=>FRAGMENT JSON zipped SIZE: "+fieldEntry.buffer().buffer().limit());
                    System.out.println("=>FRAGMENT JSON zipped SIZE: "+ bt.limit());

                    if(totalSize == 0) // not a first segment
                        totalSize = totalSizes.get(guid).longValue();
                    if (bt.limit() == totalSize) {
                        // there is only one segment, we are ready
                        // unzip using gzip
                        String strFlatFrag = unzipPayload(bt.array());
                        System.out.println("=>FRAGMENT JSON STRING: " + strFlatFrag);
                    } else {
                        ArrayList<ByteBuffer> alFrags;
                        if(!fragBuilderHash.containsKey(guid)) {
                            alFrags = new ArrayList<ByteBuffer>();
                            // for Deep Clone
                            // alFrags.add(fieldEntry.buffer().buffer());
                            alFrags.add(bt);

                            fragBuilderHash.put(guid, alFrags);
                        } else {
                            alFrags = fragBuilderHash.get(guid);
                            System.out.println("=>FRAGMENT assembled alFrags guid: " + guid);
                            // for Deep Clone
                            // alFrags.add(fieldEntry.buffer().buffer());
                            alFrags.add(bt);

                            int sizeNow = 0;
                            for(int i = 0; i < alFrags.size(); i++) {
                                sizeNow+=alFrags.get(i).limit();
                                System.out.println("=>FRAGMENT alFrags.get(" + i +").limit() SIZE: " + alFrags.get(i).limit());
                            }
                            System.out.println("=>FRAGMENT assembled sizeNow SIZE: " + sizeNow);
                            if(sizeNow < totalSize) {
                                fragBuilderHash.put(guid, alFrags);
                            }  else if(sizeNow == totalSize){
                                // we are ready
                                String strFlatFrag = "";
                                ByteArrayOutputStream fragOutputStream = new ByteArrayOutputStream( );

                                for(int i = 0; i < alFrags.size(); i++) {
                                    try {
                                        fragOutputStream.write( alFrags.get(i).array(),0, alFrags.get(i).limit() );
                                    } catch(Exception e) {}
                                }
                                strFlatFrag+= unzipPayload(fragOutputStream.toByteArray());
                                System.out.println("=>FRAGMENT JSON STRING: " + strFlatFrag);
                                fragBuilderHash.remove(guid);
                            } else {
                                System.out.println("This is a size "+ sizeNow);
                                System.exit(-1);
                            }

                        }
                    }
                }
            } else if(fieldEntry.fieldId() == FID_TOT_SIZE) {
                totalSize = fieldEntry.uintValue();
                System.out.println("=>totalSize SIZE: " + totalSize);
            } else if(fieldEntry.fieldId() == FID_GUID) {
                guid = fieldEntry.rmtes().toString();
                if(totalSize!=0)
                    totalSizes.put(guid, (new Long(totalSize)));
            }
        }
    }

    public String unzipPayload(byte[] bytes) {

        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        ByteArrayInputStream bis = new ByteArrayInputStream(bytes);
        try {
            GZIPInputStream gis = new GZIPInputStream(bis);
            byte[] buffer = new byte[bytes.length+1/*1024*/];
            int len = -1;
            while ((len = gis.read(buffer, 0, /*1024*/bytes.length+1)) != -1) {
                bout.write(buffer, 0, len);
            }
        } catch (Exception e) {
            System.out.println("Parse News error:" + e.getMessage());
        }
        return bout.toString();

    }


    // uncompressData
    // uncompress a complete NTA item
    public String uncompressData(byte[] bytes)
    {
        StringBuffer result = new StringBuffer();
        try {
            GZIPInputStream gis = new GZIPInputStream(new ByteArrayInputStream(bytes));
            BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(gis, "UTF-8"));

            String line = "";
            while ((line = bufferedReader.readLine()) != null) {
                result.append(line);
            }
        } catch (Exception e) {
            e.printStackTrace();
            System.out.println("Decompress Failed : " + e.getMessage() + "\n");
        }
        return result.toString();
    }


    // run
    // Connect to our provider, request for data and run the application for a short time to capture the realtime data.
    private void run()
    {
        try
        {
            // The consumer configuration class allows the customization of the consumer (OmmConsumer) interface, i.e.
            // name of the server or IP, port, connection type, login username, etc.
            consumer  = EmaFactory.createOmmConsumer(EmaFactory.createOmmConsumerConfig().consumerName("Consumer_1"));

            // Create an instance of our consumer in preparation for communication.
            // When instructing the consumer to communicate with a Provider, we provide the following
            // connection details.  The connection details are set within our configuration (OmmConsumerConfig).
            //
            // Note: 	After creating our consumer, by default, EMA will launch a background thread to manage
            //			communication and message dispatching.  Within EMA, this is controlled by the
            //			OperationModel available within the OmmConsumerConfig interface.  Refer to the
            //			documentation for more details.
            //
            //			For example, EMA sets the following model by default:
            //				config.operationModel(OmmConsumerConfig.OperationModel.API_DISPATCH);


            // Define what data we want and register it with our consumer
            //
            // When registering, we also need to specify where we want to capture all the events generated by
            // this request.
            OmmConsumerClient eventHandler = this;

            consumer.registerClient(
                    EmaFactory.createReqMsg()
                            .domainType(EmaRdm.MMT_NEWS_TEXT_ANALYTICS)
                            .serviceName("ELEKTRON_DD")
                            .name("MRN_STORY"), //MRN_TRNA
                    eventHandler
            );

            // Because we are using the default operation model: API_DISPATCH where EMA manages the dispatching of
            // events within a background thread, we must block and wait on our application thread for a short
            // period to capture our market data events.
            while(true) {
                //10 Minutes
                Thread.sleep(600000);
            }
        }
        catch (InterruptedException | OmmException excp)
        {
            System.out.println("Exception thrown: " + excp.getMessage());
        }
        finally
        {
            // Cleanup resources before we return
            if (consumer != null)
            {
                // Important we uninitialize our consumer to properly clean up, unregister, disconnect and stop
                // the background thread managing event dispatching.
                consumer.uninitialize();
                consumer = null;
            }
        }
    }

    //*************************************************************************************************
    // main
    // Application entry point to manage the life of the consumer.
    //
    //*************************************************************************************************
    public static void main(String[] args)
    {
        // Main interface to manage all consumer communication to our Provider
        basicConsumer basicConsumer = new basicConsumer();

        // Run the application for short time to manage communication to our Provider
        basicConsumer.run();

        // Done processing
        System.out.print("Processing complete");
    }


    // store fragments for re-assembly, keyed by GUID
    Hashtable<String,ArrayList<ByteBuffer>> fragBuilderHash =
            new Hashtable<String,ArrayList<ByteBuffer>>();
    // store total sizes
    Hashtable<String,Long> totalSizes = new Hashtable<String,Long>();

    final int FID_GUID 		= 4271;
    final int FID_MRN_TYPE 	= 8593;
    final int FID_MRN_SRC 	= 12215;
    final int FID_FRAG_NUM 	= 32479;
    final int FID_TOT_SIZE 	= 32480;
    final int FID_FRAGMENT 	= 32641;

}
